@@ -25,3 +25,23 @@ require("busted")
 
 require("non_strict")
 require("class")
+local say = require("say")
+local assert = require("luassert")
+
+function permanent()
+  return function()
+    return {}
+  end
+end
+
+function values()
+  return function()
+    return nil
+  end
+end
+
+function assertion_matches(_, arguments)
+  return string.match(arguments[1], arguments[2]) ~= nil
+end
+say:set("assertion.matches", "Expected substring fail.\n<String>: %s\n<Pattern>:%s")
+assert:register("assertion", "matches", assertion_matches, "assertion.matches")
